%%%-------------------------------------------------------------------
%%% String implemented over an Erlang binary.
%%%-------------------------------------------------------------------
-module(bstr).
-author('Juan Jose Comellas <jcomellas@novamens.com>').

-include_lib("eunit/include/eunit.hrl").
-export([len/1, equal/2, concat/2, nth/2, index/2, rindex/2,
         member/2, prefix/2, suffix/2,
         is_alpha/1, is_alnum/1, is_lower/1, is_upper/1, is_digit/1,
         is_xdigit/1, is_blank/1, is_space/1, insert/3, duplicate/2,
         substr/2, substr/3, left/2, left/3, right/2, right/3,
         pad/2, pad/3, lpad/2, lpad/3, rpad/2, rpad/3,
         strip/1, strip/2, lstrip/1, lstrip/2, rstrip/1, rstrip/2, chomp/1,
         split/2, join/1, join/2, lower/1, upper/1, bstr/1,
         from_atom/1, to_atom/1, to_existing_atom/1, from_list/1, to_list/1,
         from_integer/1, from_integer/2, to_integer/1, to_integer/2,
         urlencode/1, urldecode/1, xmlencode/1, xmldecode/1]).


%%--------------------------------------------------------------------
%% len(bstr()) -> integer()
%%  Return the length of a string.
%%--------------------------------------------------------------------
len(Str) when is_binary(Str) ->
    size(Str).


%%--------------------------------------------------------------------
%% equal(bstr(), bstr()) -> true | false
%%  Test if 2 strings are equal.
%%--------------------------------------------------------------------
equal(Str, Str) when is_binary(Str) ->
    true;
equal(Str, _) when is_binary(Str) ->
    false.


%%--------------------------------------------------------------------
%% concat(bstr(), bstr()) -> bstr()
%%  Concatenate 2 strings.
%%--------------------------------------------------------------------
concat(Str1, Str2) when is_binary(Str1), is_binary(Str2) ->
    <<Str1/binary, Str2/binary>>.


%%--------------------------------------------------------------------
%% nth(bstr(), integer()) -> char()
%%  Return the character in the nth position of the string.
%%--------------------------------------------------------------------
nth(Str, Pos) when is_binary(Str), Pos > 0, Pos =< size(Str) ->
    N = Pos - 1,
    <<_Head:N/binary, Char:8, _Tail/binary>> = Str,
    Char.


%%--------------------------------------------------------------------
%% index(bstr(), char()) -> integer()
%%  Return the first index of the character in a string.
%%--------------------------------------------------------------------
index(Str, Char) when is_binary(Str), is_integer(Char) ->
    index(Str, Char, 0).
index(Str, _Char, N) when N > size(Str) ->
    0;
index(Str, Char, N) ->
    case Str of
        <<_Head:N/binary, Char, _Tail/binary>> ->
            N + 1;
        <<>> ->
            0;
        _ ->
            index(Str, Char, N + 1)
    end.


%%--------------------------------------------------------------------
%% rindex(bstr(), char()) -> integer()
%%  Return the last index of the character in a string.
%%--------------------------------------------------------------------
rindex(Str, Char) when is_binary(Str), is_integer(Char) ->
    rindex(Str, Char, size(Str) - 1, 0).
rindex(_Str, _Char, Before, _After) when Before < 0 ->
    0;
rindex(Str, Char, Before, After) ->
    case Str of
        <<_Head:Before/binary, Char, _Tail:After/binary>> ->
            Before + 1;
        % We should never get an empty bstr, as we're covering that in the
        % guard of the function, but we check for it just in case.
        <<>> ->
            erlang:error(badarg);
        _ ->
            rindex(Str, Char, Before - 1, After + 1)
    end.


%%--------------------------------------------------------------------
%% member(bstr(), char()) -> bool()
%%  Return whether the character is present in the string.
%%--------------------------------------------------------------------
member(Str, Char) ->
    index(Str, Char) > 0.


%%--------------------------------------------------------------------
%% prefix(bstr(), bstr()) -> bool()
%%  Indicates whether a string is a prefix of another one.
%%--------------------------------------------------------------------
prefix(Str, Prefix) when is_binary(Str), is_binary(Prefix) ->
    N = size(Prefix),
    case Str of
        <<Prefix:N/binary, _Tail/binary>> ->
            true;
        _ ->
            false
    end.


%%--------------------------------------------------------------------
%% suffix(bstr(), bstr()) -> bool()
%%  Indicates whether a string is a suffix of another one.
%%--------------------------------------------------------------------
suffix(Str, Suffix) when is_binary(Str), is_binary(Suffix) ->
    N = size(Str) - size(Suffix),
    case Str of
        <<_Head:N/binary, Suffix/binary>> ->
            true;
        _ ->
            false
    end.


%%--------------------------------------------------------------------
%% is_alpha(binary()) -> boolean()
%%  Determines if a string is composed of alphabetic characters.
%%--------------------------------------------------------------------
is_alpha(Str) when is_binary(Str) ->
    is_x(Str, fun is_alpha_char/1, 0).

%% Determine if a character is alphabetic.
is_alpha_char(Char) ->
    ((Char >= $A) and (Char =< $Z)) or
    ((Char >= $a) and (Char =< $z)).


%%--------------------------------------------------------------------
%% is_alnum(binary()) -> boolean()
%%  Determines if a string is composed of alphanumeric characters.
%%--------------------------------------------------------------------
is_alnum(Str) when is_binary(Str) ->
    is_x(Str, fun is_alnum_char/1, 0).

%% Determine if a character is alphanumeric.
is_alnum_char(Char) ->
    ((Char >= $A) and (Char =< $Z)) or
    ((Char >= $a) and (Char =< $z)) or
    ((Char >= $0) and (Char =< $9)).


%%--------------------------------------------------------------------
%% is_lower(binary()) -> boolean()
%%  Determines if a string is composed of lower-case alphabetic characters.
%%--------------------------------------------------------------------
is_lower(Str) when is_binary(Str) ->
    is_x(Str, fun is_lower_char/1, 0).

%% Determine if a character is lower-case.
is_lower_char(Char) ->
    ((Char >= $a) and (Char =< $z)).


%%--------------------------------------------------------------------
%% is_upper(binary()) -> boolean()
%%  Determines if a string is composed of upper-case alphabetic characters.
%%--------------------------------------------------------------------
is_upper(Str) when is_binary(Str) ->
    is_x(Str, fun is_upper_char/1, 0).

%% Determine if a character is upper-case.
is_upper_char(Char) ->
    ((Char >= $A) and (Char =< $Z)).


%%--------------------------------------------------------------------
%% is_digit(binary()) -> boolean()
%%  Determines if a string is composed of digits.
%%--------------------------------------------------------------------
is_digit(Str) when is_binary(Str) ->
    is_x(Str, fun is_digit_char/1, 0).

%% Determine if a character is a digit.
is_digit_char(Char) ->
    ((Char >= $0) and (Char =< $9)).


%%--------------------------------------------------------------------
%% is_xdigit(binary()) -> boolean()
%%  Determines if a string is composed of hexadecimal digits.
%%--------------------------------------------------------------------
is_xdigit(Str) when is_binary(Str) ->
    is_x(Str, fun is_xdigit_char/1, 0).

%% Determine if a character is a digit.
is_xdigit_char(Char) ->
    ((Char >= $0) and (Char =< $9)) or
    ((Char >= $A) and (Char =< $F)) or
    ((Char >= $a) and (Char =< $f)).


%%--------------------------------------------------------------------
%% is_blank(binary()) -> boolean()
%%  Determines if a string is composed of blank characters.
%%--------------------------------------------------------------------
is_blank(Str) when is_binary(Str) ->
    is_x(Str, fun is_blank_char/1, 0).

%% Determine if a character is blank (\n, \r, \t, \f, \v).
is_blank_char(Char) ->
    ((Char =:= $\s) or (Char =:= $\t)).


%%--------------------------------------------------------------------
%% is_space(binary()) -> boolean()
%%  Determines if a string is composed of spaces or tabs.
%%--------------------------------------------------------------------
is_space(Str) when is_binary(Str) ->
    is_x(Str, fun is_space_char/1, 0).

%% Determine if a character is a space or a tab.
is_space_char(Char) ->
    ((Char =:= $\s) or (Char =:= $\n) or (Char =:= $\r) or
     (Char =:= $\t) or (Char =:= $\f) or (Char =:= $\v)).


%%--------------------------------------------------------------------
%% is_x(Str::bstr(), Fun::fun(), Offset::integer()) -> boolean()
%%  Helper function used to check whether all the characters in a string
%%  meet a specific criteria that is passed as a function to it.
%%--------------------------------------------------------------------
is_x(<<>>, _Fun, _Offset) ->
    false;
is_x(Str, Fun, Offset) ->
    case Str of
        <<_Head:Offset/binary, Char, _Tail/binary>> ->
            case Fun(Char) of
                true ->
                    is_x(Str, Fun, Offset + 1);
                false ->
                    false
            end;
        % If we reach the end we have a string composed entirely of characters
        % that meet the criteria specified in the Fun().
        _ ->
            true
    end.


%%--------------------------------------------------------------------
%% insert(bstr(), integer(), bstr()) -> bstr()
%%  Insert a string into another one at the indicated position.
%%--------------------------------------------------------------------
insert(Str, Pos, Str1) when is_binary(Str), is_integer(Pos) ->
    N = Pos - 1,
    case Str of
        <<Head:N/binary, Tail/binary>> ->
            <<Head/binary, Str1/binary, Tail/binary>>;
        _ ->
            erlang:error(badarg)
    end.


%%--------------------------------------------------------------------
%% duplicate(bstr(), Number) -> bstr()
%%  Return 'Number' copies of a string
%%--------------------------------------------------------------------
duplicate(Str, Num) when is_binary(Str), Num >= 0 ->
    duplicate_str(Str, Num, <<>>);
duplicate(Char, Num) when is_integer(Char), Num >= 0 ->
    duplicate_char(Char, Num, <<>>).
%% Make 'Num' copies of a string
duplicate_str(_Str, 0, Acc) ->
    Acc;
duplicate_str(Str, Num, Acc) ->
    duplicate_str(Str, Num - 1, <<Acc/binary, Str/binary>>).
%% Make 'Num' copies of a character
duplicate_char(_Str, 0, Acc) ->
    Acc;
duplicate_char(Char, Num, Acc) ->
    duplicate_char(Char, Num - 1, <<Acc/binary, Char>>).


%%--------------------------------------------------------------------
%% substr(bstr(), Pos) -> bstr()
%%  Return a substring starting at position 'Pos'.
%%--------------------------------------------------------------------
substr(Str, 1) when is_binary(Str) ->
    Str;
substr(Str, Pos) when is_binary(Str) ->
    N = Pos - 1,
    case Str of
        <<_Head:N/binary, Substr/binary>> ->
            Substr;
        _ ->
            <<>>
    end.
%%--------------------------------------------------------------------
%% substr(bstr(), Pos::integer(), Len::integer()) -> bstr()
%%  Return a substring starting at position 'Pos' with a length of 'Len' bytes.
%%--------------------------------------------------------------------
substr(Str, 1, Len) when is_binary(Str), Len =:= size(Str) ->
    Str;
substr(Str, Pos, Len) when is_binary(Str) ->
    N = Pos - 1,
    case Str of
        <<_Head:N/binary, Substr:Len/binary, _Tail/binary>> ->
            Substr;
        <<_Head:N/binary, Substr/binary>> ->
            Substr;
        _ ->
            <<>>
    end.


%%--------------------------------------------------------------------
%% left(bstr(), Len) -> bstr()
%%  Return a substring of 'Len' bytes starting from the beginning of the string.
%%  If the string does not have enough characters, the returned string is padded
%%  with spaces at the end to reach the desired length.
%%--------------------------------------------------------------------
left(Str, Len) when is_binary(Str), Len >= 0 ->
    left(Str, Len, $\s).
%%--------------------------------------------------------------------
%% left(bstr(), Len, Char) -> bstr()
%%  Return a substring of 'Len' bytes starting from the beginning of the string.
%%  If the string does not have enough characters, the returned string is padded
%%  with 'Chars' at the end to reach the desired length.
%%--------------------------------------------------------------------
left(Str, Len, Char) when is_binary(Str), Len >= 0, is_integer(Char) ->
    Slen = size(Str),
    if
        Slen > Len ->
            substr(Str, 1, Len);
        Slen < Len ->
            lpad(Str, Len, Char);
        Slen =:= Len
            -> Str
    end.


%%--------------------------------------------------------------------
%% right(bstr(), Len) -> bstr()
%%  Return a substring of 'Len' bytes starting from the end of the string.
%%  If the string does not have enough characters, the returned string is padded
%%  with spaces at the beginning to reach the desired length.
%%--------------------------------------------------------------------
right(Str, Len) when is_binary(Str), Len >= 0 ->
    right(Str, Len, $\s).
%%--------------------------------------------------------------------
%% right(bstr(), Len, Char) -> bstr()
%%  Return a substring of 'Len' bytes starting from the end of the string.
%%  If the string does not have enough characters, the returned string is padded
%%  with 'Chars' at the beginning to reach the desired length.
%%--------------------------------------------------------------------
right(Str, Len, Char) when is_integer(Char) ->
    Slen = size(Str),
    if
        Slen > Len ->
            substr(Str, Slen - Len + 1);
        Slen < Len ->
            rpad(Str, Len, Char);
        Slen =:= Len ->
            Str
    end.


%%--------------------------------------------------------------------
%% pad(bstr(), Len) -> bstr()
%%  Return a string of 'Len' bytes padded with spaces to the left and to the right.
%%--------------------------------------------------------------------
pad(Str, Len) when is_binary(Str), Len >= 0 ->
    pad(Str, Len, $\s).
%%--------------------------------------------------------------------
%% pad(bstr(), Len, Char) -> bstr()
%%  Return a string of 'Len' bytes padded with 'Chars' to the left and to the right.
%%--------------------------------------------------------------------
pad(Str, Len, Char) when is_binary(Str), Len >= 0 ->
    PadLen = Len - size(Str),
    if
        PadLen > 0 ->
            LeftPadLen = PadLen div 2,
            RightPadLen = PadLen - LeftPadLen,
            Padding = duplicate(Char, LeftPadLen),
            if
                RightPadLen > 0 ->
                    if
                        LeftPadLen =:= RightPadLen ->
                            <<Padding/binary, Str/binary, Padding/binary>>;
                        true ->
                            <<Padding/binary, Str/binary, Padding/binary, Char>>
                    end;
                true ->
                    <<Padding/binary, Str/binary>>
            end;
        true ->
            Str
    end.


%%--------------------------------------------------------------------
%% lpad(bstr(), Len) -> bstr()
%%  Return a string of 'Len' bytes left-padded with spaces.
%%--------------------------------------------------------------------
lpad(Str, Len) when is_binary(Str), Len >= 0 ->
    lpad(Str, Len, $\s).
%%--------------------------------------------------------------------
%% lpad(bstr(), Len, Char) -> bstr()
%%  Return a string of 'Len' bytes left-padded with 'Chars'.
%%--------------------------------------------------------------------
lpad(Str, Len, Char) when is_binary(Str), Len >= 0 ->
    PadLen = Len - size(Str),
    if
        PadLen > 0 ->
            Padding = duplicate(Char, PadLen),
            <<Padding/binary, Str/binary>>;
        true ->
            Str
    end.


%%--------------------------------------------------------------------
%% rpad(bstr(), Len) -> bstr()
%%  Return a string of 'Len' bytes right-padded with spaces.
%%--------------------------------------------------------------------
rpad(Str, Len) when is_binary(Str), Len >= 0 ->
    rpad(Str, Len, $\s).
%%--------------------------------------------------------------------
%% rpad(bstr(), Len, Char) -> bstr()
%%  Return a string of 'Len' bytes right-padded with 'Chars'.
%%--------------------------------------------------------------------
rpad(Str, Len, Char) ->
    PadLen = Len - size(Str),
    if
        PadLen > 0 ->
            Padding = duplicate(Char, PadLen),
            <<Str/binary, Padding/binary>>;
        true ->
            Str
    end.


%%--------------------------------------------------------------------
%% strip(bstr()) -> bstr()
%%  Remove all the spaces present both to the left and to the right of the string.
%%--------------------------------------------------------------------
strip(Str) when is_binary(Str) ->
    strip(Str, $\s).
%%--------------------------------------------------------------------
%% strip(bstr(), Char) -> bstr()
%%  Remove all the 'Chars' present both to the left and to the right of the string.
%%--------------------------------------------------------------------
strip(Str, Char) when is_binary(Str), is_integer(Char) ->
    rstrip(lstrip(Str, Char), Char).


%%--------------------------------------------------------------------
%% lstrip(bstr()) -> bstr()
%%  Remove all the spaces present to the left of the string.
%%--------------------------------------------------------------------
lstrip(Str) when is_binary(Str) ->
    lstrip(Str, $\s).
%%--------------------------------------------------------------------
%% lstrip(bstr(), Char) -> bstr()
%%  Remove all the 'Chars' present to the left of the string.
%%--------------------------------------------------------------------
lstrip(Str, Char) when is_binary(Str), is_integer(Char) ->
    lstrip(Str, Char, 0).
lstrip(Str, Char, Pos) ->
    case Str of
        <<_Head:Pos/binary, Char, _Tail/binary>> ->
            lstrip(Str, Char, Pos + 1);
        <<_Head:Pos/binary, Tail/binary>> ->
            Tail;
        _ ->
            <<>>
    end.


%%--------------------------------------------------------------------
%% rstrip(bstr()) -> bstr()
%%  Remove all the spaces present to the right of the string.
%%--------------------------------------------------------------------
rstrip(Str) when is_binary(Str) ->
    rstrip(Str, $\s).
%%--------------------------------------------------------------------
%% rstrip(bstr(), Char) -> bstr()
%%  Remove all the 'Chars' present to the right of the string.
%%--------------------------------------------------------------------
rstrip(Str, Char) when is_binary(Str), is_integer(Char) ->
    rstrip(Str, Char, size(Str) - 1).
rstrip(Str, Char, Pos) ->
    case Str of
        <<_Head:Pos/binary, Char, _Tail/binary>> ->
            rstrip(Str, Char, Pos - 1);
        <<_Head:Pos/binary, _Tail/binary>> ->
            N = Pos + 1,
            <<Stripped:N/binary, _Dummy/binary>> = Str,
            Stripped;
        _ ->
            <<>>
    end.


%%--------------------------------------------------------------------
%% chomp(bstr()) -> bstr()
%%  Remove all the newlines (\r and \n) present at the end of the string.
%%--------------------------------------------------------------------
chomp(Str) when is_binary(Str) ->
    chomp(Str, size(Str) - 1).
chomp(Str, Pos) ->
    case Str of
        <<_Head:Pos/binary, $\n, _Tail/binary>> ->
            chomp(Str, Pos - 1);
        <<_Head:Pos/binary, $\r, _Tail/binary>> ->
            chomp(Str, Pos - 1);
        <<_Head:Pos/binary, _Tail/binary>> ->
            N = Pos + 1,
            <<Stripped:N/binary, _Whitespace/binary>> = Str,
            Stripped;
        _ ->
            <<>>
    end.


%%--------------------------------------------------------------------
%% split(bstr(), Sep::char() | Sep::bstr()) -> list(bstr())
%%  Divide a string into a list of tokens that were originally separated
%%  by the character 'Sep'.
%%--------------------------------------------------------------------
split(<<>>, _Sep) ->
    [];
split(Str, Sep) when is_binary(Str), is_integer(Sep) ->
    lists:reverse(split_char_sep(Str, Sep, 0, []));
split(Str, Sep) when is_binary(Str), is_binary(Sep) ->
    case Sep of
        <<Char:8>> ->
            Tokens = split_char_sep(Str, Char, 0, []);
        _ ->
            Tokens = split_str_sep(Str, Sep, 0, [])
    end,
    lists:reverse(Tokens).
%% Helper function used to tokenize a string when the separator is a character.
split_char_sep(Str, Sep, Pos, Tokens) ->
    case Str of
        <<Token:Pos/binary, Sep, Tail/binary>> ->
            split_char_sep(Tail, Sep, 0, [Token|Tokens]);
        <<_Head:Pos/binary, _Tail/binary>> ->
            split_char_sep(Str, Sep, Pos + 1, Tokens);
        _ ->
            [Str|Tokens]
    end.
%% Helper function used to tokenize a string when there are multiple separators.
split_str_sep(Str, Sep, Pos, Tokens) ->
    case Str of
        <<Token:Pos/binary, Char:8, Tail/binary>> ->
            Index = index(Sep, Char),
            if
                Index > 0 ->
                    split_str_sep(Tail, Sep, 0, [Token|Tokens]);
                true ->
                    split_str_sep(Str, Sep, Pos + 1, Tokens)
            end;
        _ ->
            [Str|Tokens]
    end.


%%--------------------------------------------------------------------
%% join(list(bstr())) -> bstr()
%%  Join a a list of strings into one string.
%%--------------------------------------------------------------------
join(List) when is_list(List) ->
    join_list(List, <<>>).
join_list([Head|Tail], Acc) ->
    Value = bstr(Head),
    join_list(Tail, <<Acc/binary, Value/binary>>);
join_list([], Acc) ->
    Acc.


%%--------------------------------------------------------------------
%% join(list(bstr()), Sep::char() | Sep::bstr()) -> bstr()
%%  Join a a list of strings into one string, adding a separator
%%  between each string.
%%--------------------------------------------------------------------
join([Head|Tail], Sep) ->
    join_list(Tail, Sep, bstr(Head));
join([], _Sep) ->
    <<>>.
%% Join list using a bstr separator
join_list([Head|Tail], Sep, Acc) when is_binary(Sep) ->
    Value = bstr(Head),
    join_list(Tail, Sep, <<Acc/binary, Sep/binary, Value/binary>>);
%% Join list using a character separator
join_list([Head|Tail], Sep, Acc) when is_integer(Sep) ->
    Value = bstr(Head),
    join_list(Tail, Sep, <<Acc/binary, Sep, Value/binary>>);
join_list([], _Sep, Acc) ->
    Acc.


%%--------------------------------------------------------------------
%% lower(bstr()) -> bstr()
%%  Convert all the characters in a bstr to lowercase.
%%--------------------------------------------------------------------
lower(Str) when is_binary(Str) ->
    lower(Str, 0).
lower(Str, N) ->
    case Str of
        <<Head:N/binary, Char, _Tail/binary>> ->
            case is_upper_char(Char) of
                true ->
                    Lower = Char - $A + $a,
                    lower(Str, N + 1, <<Head/binary, Lower>>);
                false ->
                    lower(Str, N + 1)
            end;
        _ ->
            Str
    end.
lower(Str, N, Acc) ->
    case Str of
        <<_Head:N/binary, Char, _Tail/binary>> ->
            case is_upper_char(Char) of
                true ->
                    Lower = Char - $A + $a,
                    lower(Str, N + 1, <<Acc/binary, Lower>>);
                false ->
                    lower(Str, N + 1, <<Acc/binary, Char>>)
            end;
        _ ->
            Acc
    end.


%%--------------------------------------------------------------------
%% upper(bstr()) -> bstr()
%%  Convert all the characters in a bstr to uppercase.
%%--------------------------------------------------------------------
upper(Str) when is_binary(Str) ->
    upper(Str, 0).
upper(Str, N) ->
    case Str of
        <<Head:N/binary, Char, _Tail/binary>> ->
            case is_lower_char(Char) of
                true ->
                    Upper = Char - $a + $A,
                    upper(Str, N + 1, <<Head/binary, Upper>>);
                false ->
                    upper(Str, N + 1)
            end;
        _ ->
            Str
    end.
upper(Str, N, Acc) ->
    case Str of
        <<_Head:N/binary, Char, _Tail/binary>> ->
            case is_lower_char(Char) of
                true ->
                    Upper = Char - $a + $A,
                    upper(Str, N + 1, <<Acc/binary, Upper>>);
                false ->
                    upper(Str, N + 1, <<Acc/binary, Char>>)
            end;
        _ ->
            Acc
    end.


%%--------------------------------------------------------------------
%% bstr(binary() | atom() | list() | char()) -> bstr()
%%  Convert an "object" to a bstr.
%%--------------------------------------------------------------------
bstr(Bin) when is_binary(Bin) ->
    Bin;
bstr(Atom) when is_atom(Atom) ->
    list_to_binary(atom_to_list(Atom));
bstr(Integer) when is_integer(Integer) ->
    <<Integer>>;
bstr(List) when is_list(List) ->
    list_to_binary(List).


%%--------------------------------------------------------------------
%% from_atom(atom()) -> bstr()
%%  Convert an atom to a bstr.
%%--------------------------------------------------------------------
from_atom(Atom) when is_atom(Atom) ->
    list_to_binary(atom_to_list(Atom)).

%%--------------------------------------------------------------------
%% to_atom(bstr()) -> atom()
%%  Convert a bstr containing a string to an Erlang atom.
%%--------------------------------------------------------------------
to_atom(Str) when is_binary(Str) ->
    list_to_atom(binary_to_list(Str)).

%%--------------------------------------------------------------------
%% to_existing_atom(bstr()) -> atom()
%%  Convert a bstr containing a string to an Erlang atom only if the atom
%%  already existed (i.e. had been previously defined.
%%--------------------------------------------------------------------
to_existing_atom(Str) when is_binary(Str) ->
    list_to_existing_atom(binary_to_list(Str)).


%%--------------------------------------------------------------------
%% from_list(list()) -> bstr()
%%  Convert a list containing a string to a bstr.
%%--------------------------------------------------------------------
from_list(List) when is_list(List) ->
    list_to_binary(List).


%%--------------------------------------------------------------------
%% to_list(bstr()) -> string()
%%  Convert a bstr containing a string to an Erlang list/string.
%%--------------------------------------------------------------------
to_list(Str) when is_binary(Str) ->
    binary_to_list(Str).


%%--------------------------------------------------------------------
%% from_integer(integer()) -> bstr()
%%  Convert an integer to a bstr in base 10 format.
%%--------------------------------------------------------------------
from_integer(I) ->
    from_integer(I, 10).


%%--------------------------------------------------------------------
%% from_integer(integer(), Base::integer()) -> bstr()
%%  Convert an integer to a bstr in base 'n' format.
%%--------------------------------------------------------------------
from_integer(I, Base) when is_integer(I), is_integer(Base), Base >= 2, Base =< 1 + $Z - $A + 10 ->
    if
        I < 0 ->
            Str = from_integer(-I, Base, <<>>),
            <<$-, Str/binary>>;
       true ->
            from_integer(I, Base, <<>>)
    end;
from_integer(I, Base) ->
    erlang:fault(badarg, [I, Base]).
%% Helper function to convert an integer to a base 'n' representation.
from_integer(I0, Base, Acc0) ->
    D = I0 rem Base,
    I1 = I0 div Base,
    Acc1 =
        if
            D >= 10 ->
                Char = D - 10 + $A,
                <<Char, Acc0/binary>>;
            true ->
                Char = D + $0,
                <<Char, Acc0/binary>>
        end,
    if
        I1 =:= 0 ->
            Acc1;
        true ->
            from_integer(I1, Base, Acc1)
    end.


%%--------------------------------------------------------------------
%% to_integer(bstr()) -> integer()
%% Convert a bstr containing a string representing a decimal number
%% to an integer.
%%--------------------------------------------------------------------
to_integer(<<$-, Str/binary>>) ->
    -to_decimal_integer(Str, 0, 0);
to_integer(<<$+, Str/binary>>) ->
    to_decimal_integer(Str, 0, 0);
to_integer(<<>>) ->
    erlang:fault(badarg, [<<>>]);
to_integer(Str) when is_binary(Str) ->
    to_decimal_integer(Str, 0, 0).
%% Part of the function converts the string into a base-10 number
to_decimal_integer(Str, Offset, Acc) ->
    case Str of
        <<_Head:Offset/binary, Char:8, _Tail/binary>> when (Char >= $0) and (Char =< $9) ->
            to_decimal_integer(Str, Offset + 1, Acc * 10 + (Char - $0));
        <<_Head:Offset/binary>> ->
            Acc;
        _Number ->
            % We throw the same exception thrown by list_to_integer() if a
            % non-numeric character is found
            erlang:fault(badarg, [Str])
    end.


%%--------------------------------------------------------------------
%% to_integer(bstr(), Base::integer()) -> integer()
%%  Convert a bstr containing a string representing a positive number
%%  in the specified 'Base' to an integer. 'Base' must be an integer
%%  between 2 and 32.
%%--------------------------------------------------------------------
% Optimized version for base 10
to_integer(Str, 10) ->
    to_integer(Str);
to_integer(Str, Base) when is_integer(Base), Base >= 2, Base =< 1 + $Z - $A + 10 ->
    case Str of
        <<$-, Number/binary>> ->
            -to_base_n_integer(Number, 0, Base, 0);
        <<$+, Number/binary>> ->
            to_base_n_integer(Number, 0, Base, 0);
        <<>> ->
            erlang:fault(badarg, [<<>>]);
        Number ->
            to_base_n_integer(Number, 0, Base, 0)
    end;
to_integer(Str, Base) ->
    erlang:fault(badarg, [Str, Base]).
% Generic version for the rest of the bases
to_base_n_integer(Str, Offset, Base, Acc) ->
    case Str of
        <<_Head:Offset/binary, Char:8, _Tail/binary>> ->
            if
                (Char >= $0) and (Char =< $9) and (Char < Base + $0) ->
                    N = (Char - $0);
                (Char >= $A) and (Char =< $Z) and (Char < Base + $A) ->
                    N = 10 + (Char - $A);
                (Char >= $a) and (Char =< $z) and (Char < Base + $a) ->
                    N = 10 + (Char - $a);
                true ->
                    % We throw the same exception thrown by list_to_integer() if an
                    % invalid character is found
                    erlang:fault(badarg, [Str, Base]),
                    % To avoid compiler warning
                    N = 0
            end,
            to_base_n_integer(Str, Offset + 1, Base, Acc * Base + N);

        <<_Head:Offset/binary>> ->
            Acc;

        _Number ->
            % We throw the same exception thrown by list_to_integer() if a
            % non-numeric character is found
            erlang:fault(badarg, [Str])
    end.


%%--------------------------------------------------------------------
%% urlencode(bstr()) -> bstr()
%%  Encode a bstr using the URL-encoding scheme.
%%--------------------------------------------------------------------
urlencode(Str) when is_binary(Str) ->
    urlencode(Str, 0).
%% This part of the function iterates over the bstr without copying any data
%% and once it finds a character that has to be URL-encoded it switches to
%% the version of the function that accumulates the converted bstr.
urlencode(Str, N) ->
    case Str of
        <<Head:N/binary, Char, _Tail/binary>> ->
            case is_urlencoded(Char) of
                true ->
                    Hi = integer_to_hex((Char band 16#f0) bsr 4),
                    Lo = integer_to_hex((Char band 16#0f)),
                    urlencode(Str, N + 1, <<Head/binary, $%, Hi, Lo>>);

                false ->
                    urlencode(Str, N + 1)
            end;
        _ ->
            Str
    end.
urlencode(Str, N, Acc) ->
    case Str of
        <<_Head:N/binary, Char, _Tail/binary>> ->
            case is_urlencoded(Char) of
                true ->
                    Hi = integer_to_hex((Char band 16#f0) bsr 4),
                    Lo = integer_to_hex((Char band 16#0f)),
                    urlencode(Str, N + 1, <<Acc/binary, $%, Hi, Lo>>);

                false ->
                    urlencode(Str, N + 1, <<Acc/binary, Char>>)
            end;
        _ ->
            Acc
    end.

%% Determine whether a character has to be URL-encoded.
is_urlencoded(Char) ->
    not (((Char >= $0) and (Char =< $9)) or
         ((Char >= $A) and (Char =< $Z)) or
         ((Char >= $a) and (Char =< $z)) or
         (Char =:= $-) or (Char =:= $_) or
         (Char =:= $.) or (Char =:= $~)).

%% Convert an integer between 0 and 15 to an hexadecimal character.
integer_to_hex(N) when N >= 0 ->
    if
        N =< 9 ->
            $0 + N;
        N =< 15 ->
            $A - 10 + N;
        true ->
            erlang:error(badarg)
    end.


%%--------------------------------------------------------------------
%% urldecode(bstr()) -> bstr()
%%  Decode a bstr using the URL-encoding scheme.
%%--------------------------------------------------------------------
urldecode(Str) ->
    urldecode(Str, 0).
%% This part of the function iterates over the bstr without copying any data
%% and once it finds a character that is URL-encoded it switches to
%% the version of the function that accumulates the converted bstr.
urldecode(Str, N) ->
    case Str of
        <<Head:N/binary, $%, Hi:8, Lo:8, _Tail/binary>> ->
            Char = ((hex_to_integer(Hi) bsl 4) bor hex_to_integer(Lo)),
            urldecode(Str, N + 3, <<Head/binary, Char>>);
        <<_Head:N/binary, Char:8, _Tail/binary>> ->
            if
                Char =:= $% ->
                    erlang:error(badarg);
                true ->
                    urldecode(Str, N + 1)
            end;
        _ ->
            Str
    end.
urldecode(Str, N, Acc) ->
    case Str of
        <<_Head:N/binary, $%, Hi:8, Lo:8, _Tail/binary>> ->
            Char = ((hex_to_integer(Hi) bsl 4) bor hex_to_integer(Lo)),
            urldecode(Str, N + 3, <<Acc/binary, Char>>);
        <<_Head:N/binary, Char:8, _Tail/binary>> ->
            if
                Char =:= $% ->
                    erlang:error(badarg);
                true ->
                    urldecode(Str, N + 1, <<Acc/binary, Char>>)
            end;
        _ ->
            Acc
    end.

%% Convert an hexadecimal character to an integer. If the character is not an
%% hexadecimal character we return a 'badarg' exception.
hex_to_integer(Char) ->
    if
        (Char >= $0) and (Char =< $9) ->
            Char - $0;
        (Char >= $A) and (Char =< $F) ->
            Char - $A + 10;
        (Char >= $a) and (Char =< $f) ->
            Char - $a + 10;
        true ->
            erlang:error(badarg)
    end.



%%--------------------------------------------------------------------
%% xmlencode(bstr()) -> bstr()
%%  Encode a bstr using the XML-encoding scheme.
%%--------------------------------------------------------------------
xmlencode(Str) when is_binary(Str) ->
    xmlencode(Str, 0).
%% This part of the function iterates over the bstr without copying any data
%% and once it finds a character that has to be XML-encoded it switches to
%% the version of the function that accumulates the converted bstr.
xmlencode(Str, Offset) ->
    case Str of
        <<Head:Offset/binary, Char, _Tail/binary>> ->
            case is_xmlencoded(Char) of
                true ->
                    Encoded = xmlencode_char(Char),
                    xmlencode(Str, Offset + 1, <<Head/binary, $&, Encoded/binary, $;>>);

                false ->
                    xmlencode(Str, Offset + 1)
            end;
        _ ->
            Str
    end.
xmlencode(Str, Offset, Acc) ->
    case Str of
        <<_Head:Offset/binary, Char, _Tail/binary>> ->
            case is_xmlencoded(Char) of
                true ->
                    Encoded = xmlencode_char(Char),
                    xmlencode(Str, Offset + 1, <<Acc/binary, $&, Encoded/binary, $;>>);

                false ->
                    xmlencode(Str, Offset + 1, <<Acc/binary, Char>>)
            end;
        _ ->
            Acc
    end.

%% Determine whether a character has to be XML-encoded.
is_xmlencoded(Char) ->
    (Char =:= $&) or (Char =:= $<) or (Char =:= $>) or (Char =:= $') or (Char =:= $").

%% Encode a single character using the XML encoding scheme to create a
%% character entity reference.
xmlencode_char($&) ->
    <<"amp">>;
xmlencode_char($<) ->
    <<"lt">>;
xmlencode_char($>) ->
    <<"gt">>;
xmlencode_char($') ->
    <<"apos">>;
xmlencode_char($") ->
    <<"quot">>;
xmlencode_char(Char) ->
    <<Char>>.


%%--------------------------------------------------------------------
%% xmldecode(bstr()) -> bstr()
%%  Decode a bstr using the XML-encoding scheme to resolve any character
%%  entity reference present in the string.
%%--------------------------------------------------------------------
xmldecode(Str) ->
    xmldecode(Str, 0).
%% This part of the function iterates over the bstr without copying any data
%% and once it finds a character that is XML-encoded it switches to
%% the version of the function that accumulates the converted bstr.
xmldecode(Str, Offset) ->
    case xmldecode_char(Str, Offset) of
        % We found a character that does not need to be decoded
        {_Char, 1} ->
            xmldecode(Str, Offset + 1);
        % We found a character that needs to be encoded; we create the accumulator
        %% and call the function that copies each character to create a new string.
        {Char, Length} ->
            xmldecode(Str, Offset + Length, <<Str:Offset/binary, Char>>);
        eof ->
            Str
    end.
xmldecode(Str, Offset, Acc) ->
    case xmldecode_char(Str, Offset) of
        {Char, Length} ->
            xmldecode(Str, Offset + Length, <<Acc/binary, Char>>);
        eof ->
            Acc
    end.

%% xmldecode_char(bstr(), Offset:integer()) -> {char(), Length::integer()}
%%  Given a string and an offset, this function checks whether there is a
%%  character in that position that has to be decoded using the XML encoding
%%  scheme for character entity references and returns a tuple with the
%% decoded character and the length of the encoded substring in the original
%% string.
xmldecode_char(Str, Offset) ->
    case Str of
        <<_Head:Offset/binary, "&amp;", _Tail/binary>> ->
            {$&, 5};
        <<_Head:Offset/binary, "&lt;", _Tail/binary>> ->
            {$<, 4};
        <<_Head:Offset/binary, "&gt;", _Tail/binary>> ->
            {$>, 4};
        <<_Head:Offset/binary, "&apos;", _Tail/binary>> ->
            {$', 6};
        <<_Head:Offset/binary, "&quot;", _Tail/binary>> ->
            {$", 6};
        <<_Head:Offset/binary, Char:8, _Tail/binary>> ->
            {Char, 1};
        _ ->
           eof
    end.



%%%-------------------------------------------------------------------
%%% UNIT TESTS
%%%-------------------------------------------------------------------

%% Test for the len/1 function
len_1_test() ->
    ?assertMatch(0, len(<<>>)),
    ?assertMatch(3, len(<<"ABC">>)).

%% Test for the equal/2 function
equal_2_test() ->
    ?assert(equal(<<>>, <<>>)),
    ?assert(equal(<<"ABC123">>, <<"ABC123">>)),
    ?assertNot(equal(<<"ABC123">>, <<"abc123">>)).

%% Test for the concat/2 function
concat_2_test() ->
    ?assertMatch(<<"123">>, concat(<<>>, <<"123">>)),
    ?assertMatch(<<"123">>, concat(<<"123">>, <<>>)),
    ?assertMatch(<<"ABC123">>, concat(<<"ABC">>, <<"123">>)).

%% Test for the nth/2 function
nth_2_test() ->
    ?assertMatch($A, nth(<<"ABC">>, 1)),
    ?assertMatch($B, nth(<<"ABC">>, 2)),
    ?assertMatch($C, nth(<<"ABC">>, 3)),
    ?assertError(function_clause, nth(<<"ABC">>, 0)),
    ?assertError(function_clause, nth(<<"ABC">>, 4)),
    ?assertError(function_clause, nth(<<>>, 0)),
    ?assertError(function_clause, nth(<<>>, 1)).

%% Test for the index/2 function
index_2_test() ->
    ?assertMatch(0, index(<<>>, $A)),
    ?assertMatch(1, index(<<"ABC">>, $A)),
    ?assertMatch(2, index(<<"ABC">>, $B)),
    ?assertMatch(3, index(<<"ABC">>, $C)),
    ?assertMatch(0, index(<<"ABC">>, $D)),
    ?assertMatch(2, index(<<"ABCABC">>, $B)).

%% Test for the rindex/2 function
rindex_2_test() ->
    ?assertMatch(0, rindex(<<>>, $A)),
    ?assertMatch(1, rindex(<<"ABC">>, $A)),
    ?assertMatch(2, rindex(<<"ABC">>, $B)),
    ?assertMatch(3, rindex(<<"ABC">>, $C)),
    ?assertMatch(0, rindex(<<"ABC">>, $D)),
    ?assertMatch(5, rindex(<<"ABCABC">>, $B)),
    ?assertMatch(6, rindex(<<"ABCDEF">>, $F)).

%% Test for the member/2 function
member_2_test() ->
    ?assertNot(member(<<>>, $A)),
    ?assert(member(<<"ABC">>, $A)),
    ?assert(member(<<"ABC">>, $B)),
    ?assert(member(<<"ABC">>, $C)),
    ?assertNot(member(<<"ABC">>, $D)),
    ?assert(member(<<"ABCABC">>, $B)).

%% Test for the prefix/2 function
prefix_2_test() ->
    ?assert(prefix(<<>>, <<>>)),
    ?assertNot(prefix(<<>>, <<"ABC">>)),
    ?assert(prefix(<<"ABC">>, <<>>)),
    ?assert(prefix(<<"ABC">>, <<"AB">>)),
    ?assert(prefix(<<"ABC">>, <<"ABC">>)),
    ?assertNot(prefix(<<"AB">>, <<"ABC">>)).

%% Test for the suffix/2 function
suffix_2_test() ->
    ?assert(suffix(<<>>, <<>>)),
    ?assertNot(suffix(<<>>, <<"ABC">>)),
    ?assert(suffix(<<"ABC">>, <<>>)),
    ?assert(suffix(<<"ABC">>, <<"BC">>)),
    ?assert(suffix(<<"ABC">>, <<"ABC">>)),
    ?assertNot(suffix(<<"AB">>, <<"ABC">>)).

%% Test for the is_alpha/1 function
is_alpha_1_test() ->
    ?assertNot(is_alpha(<<>>)),
    ?assert(is_alpha(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_alpha(<<"ABC123">>)),
    ?assertNot(is_alpha(<<"ABC DEF">>)),
    ?assertNot(is_alpha(<<"A-B-C">>)).

%% Test for the is_alnum/1 function
is_alnum_1_test() ->
    ?assertNot(is_alnum(<<>>)),
    ?assert(is_alnum(<<"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_alnum(<<"ABC 123">>)),
    ?assertNot(is_alnum(<<"A-B-C">>)).

%% Test for the is_lower/1 function
is_lower_1_test() ->
    ?assertNot(is_lower(<<>>)),
    ?assert(is_lower(<<"abcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_lower(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    ?assertNot(is_lower(<<"0123456789">>)),
    ?assertNot(is_lower(<<"abc123">>)),
    ?assertNot(is_lower(<<"abc def">>)),
    ?assertNot(is_lower(<<"a-b-c">>)).

%% Test for the is_upper/1 function
is_upper_1_test() ->
    ?assertNot(is_upper(<<>>)),
    ?assert(is_upper(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    ?assertNot(is_upper(<<"abcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_upper(<<"0123456789">>)),
    ?assertNot(is_upper(<<"ABC123">>)),
    ?assertNot(is_upper(<<"ABC DEF">>)),
    ?assertNot(is_upper(<<"A-B-C">>)).

%% Test for the is_digit/1 function
is_digit_1_test() ->
    ?assertNot(is_digit(<<>>)),
    ?assert(is_digit(<<"0123456789">>)),
    ?assertNot(is_digit(<<"ABC123">>)),
    ?assertNot(is_digit(<<"abc123">>)),
    ?assertNot(is_digit(<<"123 456">>)),
    ?assertNot(is_digit(<<"1.23456">>)).

%% Test for the is_digit/1 function
is_xdigit_1_test() ->
    ?assertNot(is_xdigit(<<>>)),
    ?assert(is_xdigit(<<"0123456789ABCDEFabcdef">>)),
    ?assertNot(is_xdigit(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    ?assertNot(is_xdigit(<<"abcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_digit(<<"XYZ123">>)),
    ?assertNot(is_digit(<<"xyz123">>)),
    ?assertNot(is_digit(<<"123 ABC">>)),
    ?assertNot(is_digit(<<"A.BCDEF">>)).

%% Test for the is_blank/1 function
is_blank_1_test() ->
    ?assertNot(is_blank(<<>>)),
    ?assert(is_blank(<<"\s\t">>)),
    ?assertNot(is_blank(<<"\s\n\r\t\f\v">>)),
    ?assertNot(is_blank(<<"0123456789">>)),
    ?assertNot(is_blank(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    ?assertNot(is_blank(<<"abcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_digit(<<"  \t1\t  ">>)).

%% Test for the is_space/1 function
is_space_1_test() ->
    ?assertNot(is_blank(<<>>)),
    ?assert(is_space(<<"\s\n\r\t\f\v">>)),
    ?assertNot(is_space(<<"0123456789">>)),
    ?assertNot(is_space(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    ?assertNot(is_space(<<"abcdefghijklmnopqrstuvwxyz">>)),
    ?assertNot(is_space(<<"  \t1\t  ">>)).

%% Test for the insert/3 function
insert_3_test() ->
    ?assertMatch(<<"123">>, insert(<<>>, 1, <<"123">>)),
    ?assertMatch(<<"123ABC">>, insert(<<"ABC">>, 1, <<"123">>)),
    ?assertMatch(<<"A123BC">>, insert(<<"ABC">>, 2, <<"123">>)),
    ?assertMatch(<<"AB123C">>, insert(<<"ABC">>, 3, <<"123">>)),
    ?assertMatch(<<"ABC123">>, insert(<<"ABC">>, 4, <<"123">>)),
    ?assertError(badarg, insert(<<"ABC">>, 0, <<"123">>)),
    ?assertError(badarg, insert(<<"ABC">>, 5, <<"123">>)).

%% Test for the substr/2 function
substr_2_test() ->
    ?assertMatch(<<>>, substr(<<>>, 2)),
    ?assertMatch(<<"ABC">>, substr(<<"ABC">>, 1)),
    ?assertMatch(<<"BC">>, substr(<<"ABC">>, 2)),
    ?assertMatch(<<"C">>, substr(<<"ABC">>, 3)),
    ?assertMatch(<<>>, substr(<<"ABC">>, 4)).

%% Test for the substr/3 function
substr_3_test() ->
    ?assertMatch(<<>>, substr(<<>>, 1, 2)),
    ?assertMatch(<<"A">>, substr(<<"ABC">>, 1, 1)),
    ?assertMatch(<<"AB">>, substr(<<"ABC">>, 1, 2)),
    ?assertMatch(<<"ABC">>, substr(<<"ABC">>, 1, 3)),
    ?assertMatch(<<"ABC">>, substr(<<"ABC">>, 1, 4)),
    ?assertMatch(<<"B">>, substr(<<"ABC">>, 2, 1)),
    ?assertMatch(<<"BC">>, substr(<<"ABC">>, 2, 2)),
    ?assertMatch(<<"BC">>, substr(<<"ABC">>, 2, 3)),
    ?assertMatch(<<"BC">>, substr(<<"ABC">>, 2, 4)),
    ?assertMatch(<<"C">>, substr(<<"ABC">>, 3, 1)),
    ?assertMatch(<<"C">>, substr(<<"ABC">>, 3, 2)),
    ?assertMatch(<<>>, substr(<<"ABC">>, 4, 1)).

%% Test for the left/2 function
left_2_test() ->
    ?assertMatch(<<>>, left(<<>>, 0)),
    ?assertMatch(<<" ">>, left(<<>>, 1)),
    ?assertMatch(<<"AB">>, left(<<"ABC">>, 2)),
    ?assertMatch(<<"ABC">>, left(<<"ABC">>, 3)),
    ?assertMatch(<<" ABC">>, left(<<"ABC">>, 4)).

%% Test for the left/3 function
left_3_test() ->
    ?assertMatch(<<>>, left(<<>>, 0, $*)),
    ?assertMatch(<<"*">>, left(<<>>, 1, $*)),
    ?assertMatch(<<"AB">>, left(<<"ABC">>, 2, $*)),
    ?assertMatch(<<"ABC">>, left(<<"ABC">>, 3, $*)),
    ?assertMatch(<<"*ABC">>, left(<<"ABC">>, 4, $*)).

%% Test for the left/2 function
right_2_test() ->
    ?assertMatch(<<>>, right(<<>>, 0)),
    ?assertMatch(<<" ">>, right(<<>>, 1)),
    ?assertMatch(<<"BC">>, right(<<"ABC">>, 2)),
    ?assertMatch(<<"ABC">>, right(<<"ABC">>, 3)),
    ?assertMatch(<<"ABC ">>, right(<<"ABC">>, 4)).

%% Test for the left/3 function
right_3_test() ->
    ?assertMatch(<<>>, right(<<>>, 0, $*)),
    ?assertMatch(<<"*">>, right(<<>>, 1, $*)),
    ?assertMatch(<<"BC">>, right(<<"ABC">>, 2, $*)),
    ?assertMatch(<<"ABC">>, right(<<"ABC">>, 3, $*)),
    ?assertMatch(<<"ABC*">>, right(<<"ABC">>, 4, $*)).

%% Test for the pad/2 function
pad_2_test() ->
    ?assertMatch(<<>>, pad(<<>>, 0)),
    ?assertMatch(<<" ">>, pad(<<>>, 1)),
    ?assertMatch(<<"ABC">>, pad(<<"ABC">>, 2)),
    ?assertMatch(<<"ABC">>, pad(<<"ABC">>, 3)),
    ?assertMatch(<<" ABC ">>, pad(<<"ABC">>, 5)),
    ?assertMatch(<<" ABC  ">>, pad(<<"ABC">>, 6)).

%% Test for the pad/3 function
pad_3_test() ->
    ?assertMatch(<<>>, pad(<<>>, 0, $*)),
    ?assertMatch(<<"*">>, pad(<<>>, 1, $*)),
    ?assertMatch(<<"ABC">>, pad(<<"ABC">>, 2, $*)),
    ?assertMatch(<<"ABC">>, pad(<<"ABC">>, 3, $*)),
    ?assertMatch(<<"*ABC*">>, pad(<<"ABC">>, 5, $*)),
    ?assertMatch(<<"*ABC**">>, pad(<<"ABC">>, 6, $*)).

%% Test for the lpad/2 function
lpad_2_test() ->
    ?assertMatch(<<>>, lpad(<<>>, 0)),
    ?assertMatch(<<" ">>, lpad(<<>>, 1)),
    ?assertMatch(<<"ABC">>, lpad(<<"ABC">>, 2)),
    ?assertMatch(<<"ABC">>, lpad(<<"ABC">>, 3)),
    ?assertMatch(<<"  ABC">>, lpad(<<"ABC">>, 5)).

%% Test for the lpad/3 function
lpad_3_test() ->
    ?assertMatch(<<>>, lpad(<<>>, 0, $*)),
    ?assertMatch(<<"*">>, lpad(<<>>, 1, $*)),
    ?assertMatch(<<"ABC">>, lpad(<<"ABC">>, 2, $*)),
    ?assertMatch(<<"ABC">>, lpad(<<"ABC">>, 3, $*)),
    ?assertMatch(<<"**ABC">>, lpad(<<"ABC">>, 5, $*)).

%% Test for the rpad/2 function
rpad_2_test() ->
    ?assertMatch(<<>>, rpad(<<>>, 0)),
    ?assertMatch(<<" ">>, rpad(<<>>, 1)),
    ?assertMatch(<<"ABC">>, rpad(<<"ABC">>, 2)),
    ?assertMatch(<<"ABC">>, rpad(<<"ABC">>, 3)),
    ?assertMatch(<<"ABC  ">>, rpad(<<"ABC">>, 5)).

%% Test for the rpad/3 function
rpad_3_test() ->
    ?assertMatch(<<>>, rpad(<<>>, 0, $*)),
    ?assertMatch(<<"*">>, rpad(<<>>, 1, $*)),
    ?assertMatch(<<"ABC">>, rpad(<<"ABC">>, 2, $*)),
    ?assertMatch(<<"ABC">>, rpad(<<"ABC">>, 3, $*)),
    ?assertMatch(<<"ABC**">>, rpad(<<"ABC">>, 5, $*)).


%% Test for the strip/2 function
strip_2_test() ->
    ?assertMatch(<<>>, strip(<<>>)),
    ?assertMatch(<<>>, strip(<<" ">>)),
    ?assertMatch(<<"ABC">>, strip(<<"ABC  ">>)),
    ?assertMatch(<<"ABC">>, strip(<<"  ABC">>)),
    ?assertMatch(<<"ABC">>, strip(<<"  ABC  ">>)).

%% Test for the strip/3 function
strip_3_test() ->
    ?assertMatch(<<>>, strip(<<>>, $*)),
    ?assertMatch(<<>>, strip(<<"*">>, $*)),
    ?assertMatch(<<"ABC">>, strip(<<"ABC**">>, $*)),
    ?assertMatch(<<"ABC">>, strip(<<"**ABC">>, $*)),
    ?assertMatch(<<"ABC">>, strip(<<"**ABC**">>, $*)).


%% Test for the lstrip/2 function
lstrip_2_test() ->
    ?assertMatch(<<>>, lstrip(<<>>)),
    ?assertMatch(<<>>, lstrip(<<" ">>)),
    ?assertMatch(<<"ABC  ">>, lstrip(<<"ABC  ">>)),
    ?assertMatch(<<"ABC">>, lstrip(<<"  ABC">>)),
    ?assertMatch(<<"ABC  ">>, lstrip(<<"  ABC  ">>)).

%% Test for the lstrip/3 function
lstrip_3_test() ->
    ?assertMatch(<<>>, lstrip(<<>>, $*)),
    ?assertMatch(<<>>, lstrip(<<"*">>, $*)),
    ?assertMatch(<<"ABC**">>, lstrip(<<"ABC**">>, $*)),
    ?assertMatch(<<"ABC">>, lstrip(<<"**ABC">>, $*)),
    ?assertMatch(<<"ABC**">>, lstrip(<<"**ABC**">>, $*)).


%% Test for the rstrip/2 function
rstrip_2_test() ->
    ?assertMatch(<<>>, rstrip(<<>>)),
    ?assertMatch(<<>>, rstrip(<<" ">>)),
    ?assertMatch(<<"ABC">>, rstrip(<<"ABC  ">>)),
    ?assertMatch(<<"  ABC">>, rstrip(<<"  ABC">>)),
    ?assertMatch(<<"  ABC">>, rstrip(<<"  ABC  ">>)).

%% Test for the rstrip/3 function
rstrip_3_test() ->
    ?assertMatch(<<>>, rstrip(<<>>, $*)),
    ?assertMatch(<<>>, rstrip(<<"*">>, $*)),
    ?assertMatch(<<"ABC">>, rstrip(<<"ABC**">>, $*)),
    ?assertMatch(<<"**ABC">>, rstrip(<<"**ABC">>, $*)),
    ?assertMatch(<<"**ABC">>, rstrip(<<"**ABC**">>, $*)).

%% Test for the chomp/1 function
chomp_1_test() ->
    ?assertMatch(<<>>, chomp(<<>>)),
    ?assertMatch(<<>>, chomp(<<"\n">>)),
    ?assertMatch(<<"ABC">>, chomp(<<"ABC">>)),
    ?assertMatch(<<"ABC">>, chomp(<<"ABC\r\n">>)),
    ?assertMatch(<<"\r\nABC">>, chomp(<<"\r\nABC">>)),
    ?assertMatch(<<"\n\nABC">>, chomp(<<"\n\nABC\n\n">>)).

%% Test for the split/2 function
split_2_test() ->
    % Tokenize using a single separator
    ?assertMatch([], split(<<>>, $.)),
    ?assertMatch([<<"ABC">>], split(<<"ABC">>, $.)),
    ?assertMatch([<<>>, <<"ABC">>], split(<<".ABC">>, $.)),
    ?assertMatch([<<"ABC">>, <<>>], split(<<"ABC-">>, $-)),
    ?assertMatch([<<"ABC">>, <<"DEF">>], split(<<"ABC_DEF">>, $_)),
    ?assertMatch([<<"ABC">>, <<"DEF">>, <<"GHI">>], split(<<"ABC&DEF&GHI">>, $&)),
    % Tokenize using a group of separators
    ?assertMatch([<<"ABC">>], split(<<"ABC">>, <<".-_,">>)),
    ?assertMatch([<<>>, <<"ABC">>], split(<<".ABC">>, <<".-_,">>)),
    ?assertMatch([<<"ABC">>, <<>>], split(<<"ABC-">>, <<".-_,">>)),
    ?assertMatch([<<"ABC">>, <<"DEF">>], split(<<"ABC_DEF">>, <<"_">>)),
    ?assertMatch([<<"ABC">>, <<"DEF">>, <<"GHI">>], split(<<"ABC+DEF GHI">>, <<"+ ">>)).

%% Test for the join/1 function
join_1_test() ->
    % Empty elements
    ?assertMatch(<<>>, join([<<>>, <<>>])),
    ?assertMatch(<<"ABC">>, join([<<"ABC">>, <<>>, <<>>])),
    ?assertMatch(<<"123">>, join([<<>>, <<"123">>, <<>>])),
    ?assertMatch(<<"abc">>, join([<<>>, <<>>, <<"abc">>])),
    % Single element
    ?assertMatch(<<"ABC">>, join([<<"ABC">>])),
    % Mutliple elements
    ?assertMatch(<<"ABC123abc">>, join([<<"ABC">>, <<"123">>, <<"abc">>])),
    % Multiple elements with different data types
    ?assertMatch(<<"ABC123abc">>, join([<<"ABC">>, "123", abc])),
    ?assertMatch(<<"ABC123abc">>, join(['ABC', "123", <<"abc">>])),
    ?assertMatch(<<"ABC123abc">>, join(["ABC", <<"123">>, abc])).

%% Test for the join/2 function
join_2_test() ->
    % Empty elements
    ?assertMatch(<<>>, join([], $,)),
    ?assertMatch(<<"||">>, join([<<>>, <<>>, <<>>], $|)),
    ?assertMatch(<<"ABC||">>, join([<<"ABC">>, <<>>, <<>>], $|)),
    ?assertMatch(<<"##123##">>, join([<<>>, <<"123">>, <<>>], <<"##">>)),
    ?assertMatch(<<"??abc">>, join([<<>>, <<>>, <<"abc">>], $?)),
    % Single element
    ?assertMatch(<<"ABC">>, join([<<"ABC">>], $|)),
    % Mutliple elements with character separator
    ?assertMatch(<<"ABC,123,abc">>, join([<<"ABC">>, <<"123">>, <<"abc">>], $,)),
    % Mutliple elements with string separator
    ?assertMatch(<<"ABC**123**abc">>, join([<<"ABC">>, <<"123">>, <<"abc">>], <<"**">>)),
    % Multiple elements with different data types
    ?assertMatch(<<"ABC 123 abc">>, join([<<"ABC">>, "123", abc], $\s)),
    ?assertMatch(<<"ABC 123 abc">>, join(['ABC', "123", <<"abc">>], $\s)),
    ?assertMatch(<<"ABC 123 abc">>, join(["ABC", <<"123">>, abc], $\s)),
    ?assertMatch(<<"ABC 123 abc">>, join([<<"ABC">>, "123", abc], <<" ">>)),
    ?assertMatch(<<"ABC 123 abc">>, join(['ABC', "123", <<"abc">>], <<" ">>)),
    ?assertMatch(<<"ABC 123 abc">>, join(["ABC", <<"123">>, abc], <<" ">>)).

%% Test for the lower/1 function
lower_1_test() ->
    % Empty string
    ?assertMatch(<<>>, lower(<<>>)),
    % Numbers and symbols
    ?assertMatch(<<"-=+*12345">>, lower(<<"-=+*12345">>)),
    % Upper case characters
    ?assertMatch(<<"abcde">>, lower(<<"ABCDE">>)),
    % Lower case characters
    ?assertMatch(<<"abcde">>, lower(<<"abcde">>)),
    % Combined characters
    ?assertMatch(<<"abcde">>, lower(<<"AbCdE">>)).

%% Test for the upper/1 function
upper_1_test() ->
    % Empty string
    ?assertMatch(<<>>, upper(<<>>)),
    % Numbers and symbols
    ?assertMatch(<<"-=+*12345">>, upper(<<"-=+*12345">>)),
    % Upper case characters
    ?assertMatch(<<"ABCDE">>, upper(<<"ABCDE">>)),
    % Lower case characters
    ?assertMatch(<<"ABCDE">>, upper(<<"abcde">>)),
    % Combined characters
    ?assertMatch(<<"ABCDE">>, upper(<<"AbCdE">>)).

%% Test for the from_atom/1 function
from_atom_1_test() ->
    % Empty atom
    ?assertMatch(<<>>, from_atom('')),
    % Normal characters
    ?assertMatch(<<"abc@localhost">>, from_atom(abc@localhost)),
    % Symbols
    ?assertMatch(<<"ABC+-=[]/123">>, from_atom('ABC+-=[]/123')).

%% Test for the to_atom/1 function
to_atom_1_test() ->
    % Empty string
    ?assertMatch('', to_atom(<<>>)),
    % Normal string
    ?assertMatch(abc, to_atom(<<"abc">>)),
    % Symbols
    ?assertMatch('ABC+-=[]/123', to_atom(<<"ABC+-=[]/123">>)).

%% Test for the to_existing_atom/1 function
to_existing_atom_1_test() ->
    % Empty string
    ?assertMatch('', to_existing_atom(<<>>)),
    % We cannot test the failure of this function because once an atom is
    % created it cannot be removed from the virtual machine's atom table.
    % Str2 = <<"***THIS_ATOM_MUST_NOT_BE_ALREADY_DEFINED***">>,
    % ?assertError(badarg, to_existing_atom(Str2)),
    % Normal string
    Atom1 = abc,
    ?assertMatch(abc, to_existing_atom(<<"abc">>)).

%% Test for the from_list/1 function
from_list_1_test() ->
    % Empty string
    ?assertMatch(<<>>, from_list([])),
    % Normal string
    ?assertMatch(<<"ABC123">>, from_list("ABC123")).

%% Test for the to_list/1 function
to_list_1_test() ->
    % Empty string
    ?assertMatch([], to_list(<<>>)),
    % Normal string
    ?assertMatch("ABC123", to_list(<<"ABC123">>)).

%% Test for the from_integer/1 function
from_integer_1_test() ->
    % Positive integer
    ?assertMatch(<<"12345">>, from_integer(12345)),
    % Negative integer
    ?assertMatch(<<"-12345">>, from_integer(-12345)),
    % Floating point number
    ?assertError(badarg, from_integer(12345.67)).

%% Test for the to_integer/1 function
to_integer_1_test() ->
    % Positive integer
    ?assertMatch(12345, to_integer(<<"12345">>)),
    % Positive integer (explicit)
    ?assertMatch(12345, to_integer(<<"+12345">>)),
    % Negative integer
    ?assertMatch(-12345, to_integer(<<"-12345">>)),
    % Empty string
    ?assertError(badarg, to_integer(<<>>)),
    % Floating point number
    ?assertError(badarg, to_integer(<<"12345.67">>)),
    % Invalid characters
    ?assertError(badarg, to_integer(<<"ABC">>)).

%% Test for the from_integer/2 function
from_integer_2_test() ->
    % Base-2 integer
    ?assertMatch(<<"0">>, from_integer(0, 2)),
    ?assertMatch(<<"1">>, from_integer(1, 2)),
    ?assertMatch(<<"110110">>, from_integer(54, 2)),
    ?assertMatch(<<"-1000000">>, from_integer(-64, 2)),
    % Base-16 integer
    ?assertMatch(<<"A">>, from_integer(10, 16)),
    ?assertMatch(<<"D4BE">>, from_integer(54462, 16)),
    ?assertMatch(<<"-D4BE">>, from_integer(-54462, 16)),
    % Floating point number
    ?assertError(badarg, from_integer(12345.67, 16)).

%% Test for the to_integer/2 function
to_integer_2_test() ->
    % Base-2 integer
    ?assertMatch(1, to_integer(<<"1">>, 2)),
    ?assertMatch(54, to_integer(<<"110110">>, 2)),
    ?assertMatch(-54, to_integer(<<"-110110">>, 2)),
    % Base-16 integer
    ?assertMatch(10, to_integer(<<"A">>, 16)),
    ?assertMatch(54462, to_integer(<<"D4BE">>, 16)),
    ?assertMatch(54462, to_integer(<<"d4be">>, 16)),
    % Empty string
    ?assertError(badarg, to_integer(<<>>, 2)),
    % Invalid characters
    ?assertError(badarg, to_integer(<<"-ABC">>)).

%% Test for the urlencode/1 function
urlencode_1_test() ->
    % Unreserved characters (lower case letters)
    Str1 = <<"abcdefghijklmnopqrstuvwxyz">>,
    ?assertEqual(Str1, urlencode(Str1)),
    % Unreserved characters (upper case letters)
    Str2 = <<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>,
    ?assertEqual(Str2, urlencode(Str2)),
    % Unreserved characters (numbers and symbols)
    Str3 = <<"01234567890-_.~">>,
    ?assertEqual(Str3, urlencode(Str3)),
    % Reserved characters
    Str4 = <<"!*'();:@&=+$,/?%#[]">>,
    Enc4 = <<"%21%2A%27%28%29%3B%3A%40%26%3D%2B%24%2C%2F%3F%25%23%5B%5D">>,
    ?assertEqual(Enc4, urlencode(Str4)),
    % Control characters
    Str5 = <<16#00, 16#01, 16#02, 16#03, 16#04, 16#05, 16#06, 16#07,
             16#08, 16#09, 16#0A, 16#0B, 16#0C, 16#0D, 16#0E, 16#0F,
             16#10, 16#11, 16#12, 16#13, 16#14, 16#15, 16#16, 16#17,
             16#18, 16#19, 16#1A, 16#1B, 16#1C, 16#1D, 16#1E, 16#1F>>,
    Enc5 = <<"%00%01%02%03%04%05%06%07"
             "%08%09%0A%0B%0C%0D%0E%0F"
             "%10%11%12%13%14%15%16%17"
             "%18%19%1A%1B%1C%1D%1E%1F">>,
    ?assertEqual(Enc5, urlencode(Str5)),
    % Text with embedded spaces
    Str6 = <<"This is a test for urlencode">>,
    Enc6 = <<"This%20is%20a%20test%20for%20urlencode">>,
    ?assertEqual(Enc6, urlencode(Str6)).

%% Test for the urldecode/1 function
urldecode_1_test() ->
    % Unreserved characters (lower case letters)
    Str1 = <<"abcdefghijklmnopqrstuvwxyz">>,
    ?assertEqual(Str1, urldecode(Str1)),
    % Unreserved characters (upper case letters)
    Str2 = <<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>,
    ?assertEqual(Str2, urldecode(Str2)),
    % Unreserved characters (numbers and symbols)
    Str3 = <<"01234567890-_.~">>,
    ?assertEqual(Str3, urldecode(Str3)),
    % Reserved characters
    Enc4 = <<"%21%2A%27%28%29%3B%3A%40%26%3D%2B%24%2C%2F%3F%25%23%5B%5D">>,
    ?assertMatch(<<"!*'();:@&=+$,/?%#[]">>, urldecode(Enc4)),
    % Control characters
    Str5 = <<16#00, 16#01, 16#02, 16#03, 16#04, 16#05, 16#06, 16#07,
             16#08, 16#09, 16#0A, 16#0B, 16#0C, 16#0D, 16#0E, 16#0F,
             16#10, 16#11, 16#12, 16#13, 16#14, 16#15, 16#16, 16#17,
             16#18, 16#19, 16#1A, 16#1B, 16#1C, 16#1D, 16#1E, 16#1F>>,
    Enc5 = <<"%00%01%02%03%04%05%06%07"
             "%08%09%0A%0B%0C%0D%0E%0F"
             "%10%11%12%13%14%15%16%17"
             "%18%19%1A%1B%1C%1D%1E%1F">>,
    ?assertEqual(Str5, urldecode(Enc5)),
    % Text with embedded spaces
    Enc6 = <<"This%20is%20a%20test%20for%20urlencode">>,
    ?assertMatch(<<"This is a test for urlencode">>, urldecode(Enc6)),
    % Invalid length of URL-encoded character
    ?assertError(badarg, urldecode(<<"ABC%Z">>)),
    % Invalid URL-encoded character
    ?assertError(badarg, urldecode(<<"ABC%ZZ">>)).

%% Test for the xmlencode/1 function
xmlencode_1_test() ->
    % Empty string
    ?assert(xmlencode(<<>>) =:= <<>>),
    % Characters that are not encoded: lower case letters
    ?assertMatch(<<"abcdefghijklmnopqrstuvwxyz">>, xmlencode(<<"abcdefghijklmnopqrstuvwxyz">>)),
    % Characters that are not encoded: upper case letters
    ?assertMatch(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>, xmlencode(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    % Characters that are not encoded: numbers and symbols
    ?assertMatch(<<"01234567890-_.~">>, xmlencode(<<"01234567890-_.~">>)),
    % Characters that are encoded
    ?assertMatch(<<"&amp;&lt;&gt;&apos;&quot;">>, xmlencode(<<"&<>'\"">>)),
    % Text with embedded spaces
    ?assertMatch(<<"&lt;element attr=&quot;1234&quot;/&gt;">>, xmlencode(<<"<element attr=\"1234\"/>">>)).

%% Test for the xmldecode/1 function
xmldecode_1_test() ->
    % Empty string
    ?assert(xmldecode(<<>>) =:= <<>>),
    % Characters that are not encoded: lower case letters
    ?assertMatch(<<"abcdefghijklmnopqrstuvwxyz">>, xmldecode(<<"abcdefghijklmnopqrstuvwxyz">>)),
    % Characters that are not encoded: upper case letters
    ?assertMatch(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>, xmldecode(<<"ABCDEFGHIJKLMNOPQRSTUVWXYZ">>)),
    % Characters that are not encoded: numbers and symbols
    ?assertMatch(<<"01234567890-_.~">>, xmldecode(<<"01234567890-_.~">>)),
    % Characters that are encoded
    ?assertMatch(<<"&<>'\"">>, xmldecode(<<"&amp;&lt;&gt;&apos;&quot;">>)),
    % Text with embedded spaces
    ?assertMatch(<<"<element attr=\"1234\"/>">>, xmldecode(<<"&lt;element attr=&quot;1234&quot;/&gt;">>)),
    % Invalid encoding
    ?assertMatch(<<"**&abc;**">>, xmldecode(<<"**&abc;**">>)).
